@@ -4,7 +4,7 @@ from scipy.interpolate import interp1d
 from typing import Dict, List, Optional
 from pandas.api.types import is_numeric_dtype
 from pymsm.utils import stepfunc
-from pymsm.event_specific_fitter import EventSpecificFitter
+from pymsm.event_specific_fitter import EventSpecificFitter, CoxWrapper
 
 
 class EventSpecificModel:
@@ -20,28 +20,17 @@ class EventSpecificModel:
 
     failure_type: int
     model: EventSpecificFitter
-    coefficients: Optional[np.ndarray]
     unique_event_times: Optional[np.ndarray]
-    baseline_hazard: Optional[np.ndarray]
-    cumulative_baseline_hazard_function: Optional[np.ndarray]
 
     def __init__(self, failure_type=None, model=None):
         self.failure_type = failure_type
         self.model = model
-        self.coefficients = None
         self.unique_event_times = None
-        self.baseline_hazard = None
-        self.cumulative_baseline_hazard_function = None
 
     def extract_necessary_attributes(self) -> None:
         """Extract relevant arrays from event specific cox model
         """
-        self.coefficients = self.model.get_coefficients()
         self.unique_event_times = self.model.get_unique_event_times()
-        self.baseline_hazard = self.model.get_baseline_hazard()
-        self.cumulative_baseline_hazard_function = (
-            self.model.get_baseline_cumulative_hazard()
-        )
 
 
 class CompetingRisksModel:
@@ -57,7 +46,7 @@ class CompetingRisksModel:
         crm = CompetingRisksModel(CoxWrapper)
         data = create_test_data(N=1000)
         crm.fit(df=data, duration_col='T', event_col='transition', cluster_col='id')
-    
+
     Attributes
     ----------
     failure_types: list
@@ -72,7 +61,7 @@ class CompetingRisksModel:
     event_specific_models: Dict[int, EventSpecificModel]
     event_specific_fitter: EventSpecificFitter
 
-    def __init__(self, event_specific_fitter):
+    def __init__(self, event_specific_fitter=CoxWrapper):
         self.failure_types = []
         self.event_specific_models = {}
         self.event_specific_fitter = event_specific_fitter
@@ -110,7 +99,7 @@ class CompetingRisksModel:
         assert (df[duration_col] >= 0).any(), "duration column has negative values"
 
         # failure types should be integers from 0 to m, not necessarily consecutive
-        assert df[event_col].dtypes == int, "event column needs to be of type int"
+        assert (pd.api.types.is_integer_dtype(df[event_col].dtypes)), "event column needs to be of type int"
         assert (
             df[event_col].min() >= 0
         ), "Failure types need to zero or positive integers"
@@ -273,71 +262,12 @@ class CompetingRisksModel:
         np.ndarray
             hazard at unique event times
         """
-        # the hazard is given by multiplying the baseline hazard (which has value per unique event time) by the partial hazard
-        hazard = self._baseline_hazard(failure_type) * (
-            self._partial_hazard(failure_type, sample_covariates)
+        hazard = self.event_specific_models[failure_type].model.get_hazard(
+            sample_covariates
         )
         assert len(hazard) == len(self.unique_event_times(failure_type))
         return hazard
 
-    @staticmethod
-    def cumulative_baseline_hazard_step_function(
-            model: EventSpecificFitter
-    ) -> interp1d:
-        """Create interpolation step function for the cumulative baseline hazard
-
-        Parameters
-        ----------
-        model : EventSpecificFitter
-            event specific model
-
-        Returns
-        -------
-        interp1d
-            interpolation step function for the cumulative baseline hazard
-        """
-        return stepfunc(
-            model.get_unique_event_times(),
-            model.get_baseline_cumulative_hazard(),
-        )
-
-    def _baseline_hazard(self, failure_type: int) -> np.ndarray:
-        """Get baseline hazard for specific failure type
-
-        Parameters
-        ----------
-        failure_type : int
-            failure type of interest
-
-        Returns
-        -------
-        np.ndarray
-            baseline hazard for specific failure type
-        """
-        return self.event_specific_models[failure_type].baseline_hazard
-
-    def _partial_hazard(
-        self, failure_type: int, sample_covariates: np.ndarray
-    ) -> np.ndarray:
-        """Get partial hazard for specific failure type and set of covariates
-
-        Parameters
-        ----------
-        failure_type : int
-            failure type of interest
-        sample_covariates : np.ndarray
-            covariates
-
-        Returns
-        -------
-        np.ndarray
-            partial hazard for specific failure type and set of covariates
-        """
-        # simply e^x_dot_beta for the chosen failure type's coefficients
-        coefs = self.event_specific_models[failure_type].coefficients
-        x_dot_beta = np.dot(sample_covariates, coefs)
-        return np.exp(x_dot_beta)
-
     def unique_event_times(self, failure_type: int) -> np.ndarray:
         """Fetch unique event times for specific failure type
 
@@ -374,10 +304,9 @@ class CompetingRisksModel:
         exponent = np.zeros_like(t)
         for type in self.failure_types:
             exponent = exponent - (
-                self.cumulative_baseline_hazard_step_function(
-                    self.event_specific_models[type].model
-                )(t)
-                * (self._partial_hazard(type, sample_covariates))
+                self.event_specific_models[type].model.get_cumulative_hazard(
+                    t, sample_covariates
+                )
             )
         survival_function_at_t = np.exp(exponent)
         assert len(survival_function_at_t) == len(t)
@@ -496,3 +425,9 @@ class CompetingRisksModel:
             ) / self.survival_function(np.array([time_passed]), sample_covariates)
 
         return predictions
+
+    def print_summary(self):
+        """Print summary of fitted models"""
+        for event_of_interest in self.failure_types:
+            print(f"Model for failure type {event_of_interest}:\n")
+            self.event_specific_models[event_of_interest].model.print_summary()
